import math
import json
import os

import numpy as np
from numpy import genfromtxt
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

class Table(object):

    def custom_dict_to_str(self, d):
        if not d:
            return ''
        else:
            return str(d)
    
    def plot_table(self, path_to_results, path_to_tables):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ e)

        if not os.path.exists(path_to_tables):
            os.makedirs(path_to_tables)
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed
        bit_precision = data["general"]["bit_precision"]

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_list = data["honest_nodes"]["learning_rate"]
        momentum_list = data["honest_nodes"]["momentum"]
        wd_list = data["honest_nodes"]["weight_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_list, float):
            lr_list = [lr_list]
        if isinstance(momentum_list, float):
            momentum_list = [momentum_list]
        if isinstance(wd_list, float):
            wd_list = [wd_list]
        
        nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))

        for nb_nodes in nb_workers:
            for nb_byzantine in nb_byz:
                for lr in lr_list:
                    for momentum in momentum_list:
                        for wd in wd_list:
                            for data_dist in data_distributions:
                                alpha = ""
                                if data_dist["name"] == "dirichlet_niid":
                                    alpha = data_dist["parameters"]["alpha"]
                                
                                big_table = np.zeros((len(aggregators), len(attacks)), dtype=np.dtype('float,float'))

                                for x, agg in enumerate(aggregators):
                                    for pre_agg in pre_aggregators:
                                        if pre_agg is not None:
                                            pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
                                            pre_agg_names = "_".join(pre_agg_list_names)
                                        else:
                                            pre_agg_names = ""
                                        tab_acc = np.zeros((
                                            len(attacks), 
                                            nb_seeds, 
                                            nb_accuracies
                                        ))
                                        for i, attack in enumerate(attacks):
                                            for run in range(nb_seeds):
                                                common_name = str(dataset_name + "_" 
                                                                + model_name + "_n_" 
                                                                + str(nb_nodes) + "_f_" 
                                                                + str(nb_byzantine) + "_" 
                                                                + self.custom_dict_to_str(data_dist["name"]) 
                                                                + str(alpha) + "_" 
                                                                + self.custom_dict_to_str(agg["name"]) + "_" 
                                                                + pre_agg_names + "_" 
                                                                + self.custom_dict_to_str(attack["name"]) + "_lr_"
                                                                + str(lr) + "_mom_"
                                                                + str(momentum) + "_wd_"
                                                                + str(wd))
                                                tab_acc[i][run] = genfromtxt(path_to_results+"/"+common_name+"/train_accuracy_seed_"+str(run+seed)+".txt", delimiter=',')
                                        
                                        err = np.zeros((len(attacks), nb_accuracies))
                                        for i in range(len(err)):
                                            err[i] = (1.96*np.std(tab_acc[i], axis = 0))/math.sqrt(nb_seeds)
                                        err = np.round(err, decimals=bit_precision)
                                        
                                        for i, attack in enumerate(attacks):
                                            acc_mean = np.mean(tab_acc[i], axis=0)
                                            idx_max = np.argmax(acc_mean)
                                            acc_mean = np.round(acc_mean, decimals=bit_precision)
                                            big_table[x][i] = (acc_mean[idx_max], err[i][idx_max])
                                    
                                rows_names = [agg['name'] for agg in aggregators]
                                columns_names = [attack["name"] for attack in attacks]

                                table_name = str(dataset_name + "_"
                                                 + model_name + "_n_"
                                                 + str(nb_nodes) + "_f_"
                                                 + str(nb_byzantine) + "_"
                                                 + self.custom_dict_to_str(data_dist["name"])
                                                 + str(alpha) + "_"
                                                 + self.custom_dict_to_str(agg["name"]) + "_"
                                                 + pre_agg_names + "_lr_"
                                                 + str(lr) + "_mom_"
                                                 + str(momentum) + "_wd_"
                                                 + str(wd))
                                
                                pdf_pages = PdfPages(path_to_tables+"/"+table_name + ".pdf")

                                plt.figure(figsize=(8, 6))
                                plt.axis('off')

                                plt.table(cellText=big_table,
                                        rowLabels=rows_names,
                                        colLabels=columns_names,
                                        cellLoc= 'center',
                                        loc='center',
                                        bbox=[0.15, 0.15, 0.9, 0.9])

                                pdf_pages.savefig()
                                pdf_pages.close()

    def plot_paper(self, path_to_results, path_to_tables, path_to_hyperparameters):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_tables):
            os.makedirs(path_to_tables)
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_list = data["honest_nodes"]["learning_rate"]
        momentum_list = data["honest_nodes"]["momentum"]
        wd_list = data["honest_nodes"]["weight_decay"]
        lr_decay = data["server"]["learning_rate_decay"]
        
        bit_precision = data["general"]["bit_precision"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if not isinstance(nb_workers, list):
            nb_workers = [nb_workers]
            change = True
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_list, float):
            lr_list = [lr_list]
        if isinstance(momentum_list, float):
            momentum_list = [momentum_list]
        if isinstance(wd_list, float):
            wd_list = [wd_list]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
        for nb_nodes in nb_workers:
            for nb_byzantine in nb_byz:
                if change:
                    nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                for data_dist in data_distributions:
                    alpha_list = [""]
                    if data_dist["name"] == "dirichlet_niid":
                        alpha_list = data_dist["parameters"]["alpha"]
                    if data_dist["name"] == "gamma_similarity_niid":
                        alpha_list = data_dist["parameters"]["gamma"]
                    for alpha in alpha_list:
                        for pre_agg in pre_aggregators:
                            pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
                            pre_agg_names = "_".join(pre_agg_list_names)
                            max_worst_case_agg = np.zeros((len(aggregators), len(attacks)))
                            real_errors = np.zeros((len(aggregators), len(attacks)))
                            for k, agg in enumerate(aggregators):
                                tab_acc = np.zeros((
                                    len(attacks), 
                                    nb_seeds, 
                                    nb_accuracies
                                ))
                                hyperparameters_file_name = str(dataset_name + "_"
                                                                    + model_name + "_n_"
                                                                    + str(nb_nodes) + "_f_"
                                                                    + str(nb_byzantine) + "_"
                                                                    + self.custom_dict_to_str(data_dist["name"])
                                                                    + str(alpha) + "_"
                                                                    + pre_agg_names + "_"
                                                                    + agg["name"]
                                                                    + ".txt")
                                    
                                hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)
                                lr = hyperparameters[0]
                                momentum = hyperparameters[1]
                                if momentum == 0.0:
                                    momentum = 0
                                wd = hyperparameters[2]
                                lr_d = 1.0
                                for i, attack in enumerate(attacks):
                                    for run in range(nb_seeds):
                                        common_name = str(dataset_name + "_" 
                                                        + model_name + "_n_" 
                                                        + str(nb_nodes) + "_f_" 
                                                        + str(nb_byzantine) + "_" 
                                                        + self.custom_dict_to_str(data_dist["name"]) 
                                                        + str(alpha) + "_" 
                                                        + self.custom_dict_to_str(agg["name"]) + "_" 
                                                        + pre_agg_names + "_" 
                                                        + self.custom_dict_to_str(attack["name"]) + "_lr_"
                                                        + str(lr) + "_mom_"
                                                        + str(momentum) + "_wd_"
                                                        + str(wd)+ "_lr_decay_"
                                                        + str(lr_d))
                                        tab_acc[i][run] = genfromtxt(path_to_results+"/"+common_name+"/train_accuracy_seed_"+str(run+seed)+".txt", delimiter=',')

                                
                                err = np.zeros((len(attacks), nb_accuracies))
                                for i in range(len(err)):
                                    err[i] = (1.96*np.std(tab_acc[i], axis = 0))/math.sqrt(nb_seeds)
                                err = np.round(err, decimals=bit_precision)
                                
                                for i, attack in enumerate(attacks):
                                    acc_mean = np.mean(tab_acc[i], axis=0)
                                    idx_max = np.argmax(acc_mean)
                                    acc_mean = np.round(acc_mean, decimals=bit_precision)
                                    max_worst_case_agg[k][i] = acc_mean[idx_max]
                                    real_errors[k][i] = err[i][idx_max]

                                if not os.path.exists(path_to_hyperparameters):
                                    try:
                                        os.makedirs(path_to_hyperparameters)
                                    except OSError as error:
                                        print(f"Error creating directory: {error}")
                                
                            real_table = np.zeros((len(aggregators) + 3,len(attacks) + 1), dtype=np.dtype('float,float'))
                            row_min_values = np.array([np.inf]*len(aggregators))
                            row_min_error = np.array([np.inf]*len(aggregators))
                            column_min_values = np.array([np.inf]*len(attacks))
                            column_min_error = np.array([np.inf]*len(attacks))
                            column_max_values = np.zeros((len(attacks)))
                            column_max_error = np.zeros((len(attacks)))
                            column_avg_values = list(zip(np.mean(max_worst_case_agg, axis=0), [0.0]*(len(np.mean(max_worst_case_agg, axis=0)))))
                            
                            for i in range(len(max_worst_case_agg)):
                                for j in range(len(max_worst_case_agg[0])):
                                    real_table[i][j] = (max_worst_case_agg[i][j], real_errors[i][j])
                                    if max_worst_case_agg[i][j] < row_min_values[i]:
                                        row_min_values[i] = max_worst_case_agg[i][j]
                                        row_min_error[i] = real_errors[i][j]
                                    if max_worst_case_agg[i][j] < column_min_values[j]:
                                        column_min_values[j] = max_worst_case_agg[i][j]
                                        column_min_error[j] = real_errors[i][j]
                                    if max_worst_case_agg[i][j] > column_max_values[j]:
                                        column_max_values[j] = max_worst_case_agg[i][j]
                                        column_max_error[j] = real_errors[i][j]

                            for i in range(len(row_min_values)):
                                real_table[i][len(real_table[0])-1] = (row_min_values[i], row_min_error[i])
                            
                            for i in range(len(attacks)):
                                real_table[len(real_table)-3][i] = (column_min_values[i], column_min_error[i])
                                real_table[len(real_table)-2][i] = (column_max_values[i], column_max_error[i])
                                real_table[len(real_table)-1][i] = column_avg_values[i]

                            rows_names = [agg['name'] for agg in aggregators]
                            columns_names = [attack["name"] for attack in attacks]

                            rows_names.append("Min")
                            rows_names.append("Max")
                            rows_names.append("Average")
                            columns_names.append("Min")

                            table_name = str(dataset_name + "_"
                                            + model_name + "_n_"
                                            + str(nb_nodes) + "_f_"
                                            + str(nb_byzantine) + "_"
                                            + self.custom_dict_to_str(data_dist["name"])
                                            + str(alpha) + "_"
                                            + pre_agg_names)
                            
                            pdf_pages = PdfPages(path_to_tables + "/" + table_name + ".pdf")

                            plt.figure(figsize=(8, 6))
                            plt.axis('off')

                            plt.table(cellText=real_table,
                                    rowLabels=rows_names,
                                    colLabels=columns_names,
                                    cellLoc= 'center',
                                    loc='center',
                                    bbox=[0.15, 0.15, 0.9, 0.9])
                            

                            pdf_pages.savefig()
                            pdf_pages.close()

                            plt.close()

    def best_agg_for_pair_hetero_byz(self, path_to_results, path_to_table, path_to_hyperparameters):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_table):
            os.makedirs(path_to_table)
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if not isinstance(nb_workers, list):
            nb_workers = [nb_workers]
            change = True
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        
        data_folder = data["general"]["data_folder"]
        if data_folder is None:
            data_folder = "./data"
        lr_d = 0.1
        nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
        for data_dist in data_distributions:
            alpha_list = [""]
            if data_dist["name"] == "dirichlet_niid":
                alpha_list = data_dist["parameters"]["alpha"]
            if data_dist["name"] == "gamma_similarity_niid":
                alpha_list = data_dist["parameters"]["gamma"]
            for pre_agg in pre_aggregators:
                pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
                pre_agg_names = "_".join(pre_agg_list_names)
                cube_acc_table = np.zeros((len(aggregators), len(alpha_list), len(nb_byz)))
                cube_err_table = np.zeros((len(aggregators), len(alpha_list), len(nb_byz)))
                for a, agg in enumerate(aggregators):
                    #try:
                    acc_table = np.zeros((len(alpha_list), len(nb_byz)))
                    err_table = np.zeros((len(alpha_list), len(nb_byz)))
                    for nb_nodes in nb_workers:
                        for y, nb_byzantine in enumerate(nb_byz):
                            if change:
                                nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                            for x, alpha in enumerate(alpha_list):
                                hyperparameters_file_name = str(dataset_name + "_"
                                                                + model_name + "_n_"
                                                                + str(nb_nodes) + "_f_"
                                                                + str(nb_byzantine) + "_"
                                                                + self.custom_dict_to_str(data_dist["name"])
                                                                + str(alpha) + "_"
                                                                + pre_agg_names + "_"
                                                                + agg["name"]
                                                                + ".txt")
                                
                                hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)
                                lr = hyperparameters[0]
                                momentum = hyperparameters[1]
                                if momentum == 0.0:
                                    momentum = 0
                                wd = hyperparameters[2]
                                
                                worst_accuracy = np.inf
                                for attack in attacks:
                                    tab_acc = np.zeros((
                                            nb_seeds, 
                                            nb_accuracies
                                    ))
                                    for run in range(nb_seeds):
                                        test_accuracy_file_name = str(
                                                dataset_name + "_" 
                                                + model_name + "_" 
                                                "n_" + str(nb_nodes) + "_" 
                                                + "f_" + str(nb_byzantine) + "_" 
                                                + self.custom_dict_to_str(data_dist["name"])
                                                + str(alpha) + "_" 
                                                + self.custom_dict_to_str(agg["name"]) + "_"
                                                + pre_agg_names + "_"
                                                + self.custom_dict_to_str(attack["name"]) + "_" 
                                                + "lr_" + str(lr) + "_" 
                                                + "mom_" + str(momentum) + "_" 
                                                + "wd_" + str(wd) + "_" 
                                                + "lr_decay_" + str(lr_d) + "/"
                                            )
                                        tab_acc[run] = genfromtxt(path_to_results + "/" + test_accuracy_file_name+"/train_accuracy_seed_"+str(run+seed)+".txt", delimiter=',')
                                    
                                    err = np.zeros((nb_accuracies))
                                    err = (1.96*np.std(tab_acc, axis = 0))/math.sqrt(nb_seeds)

                                    tab_acc = tab_acc.mean(axis=0)
                                    idx_accuracy = np.argmax(tab_acc)
                                    if tab_acc[idx_accuracy] < worst_accuracy:
                                        worst_accuracy = tab_acc[idx_accuracy]
                                        error_linked = err[idx_accuracy]

                                acc_table[len(acc_table)-1-x][y] = worst_accuracy
                                err_table[len(acc_table)-1-x][y] = error_linked

                    cube_acc_table[a] = acc_table
                    cube_err_table[a] = err_table
                
                max_length = -1
                for agg in aggregators:
                    aux = len(agg["name"])
                    if aux > max_length:
                        max_length = aux
                    

                idx_better_agg = np.argmax(cube_acc_table, axis=0)
                matrix = [[0 for _ in range(len(alpha_list))] for _ in range(len(nb_byz))]
                for i in range(len(alpha_list)):
                    for j in range(len(nb_byz)):
                        best_accuracy = cube_acc_table[idx_better_agg[i][j]][i][j]
                        if best_accuracy < 0.3:
                            matrix[i][j] = "No Robust"
                        else:
                            best_agg_list = [(aggregators[idx_better_agg[i][j]]["name"] + "_" + str(np.round(best_accuracy, 4)) + "_" + str(np.round(cube_err_table[idx_better_agg[i][j]][i][j], 4)))]
                            minimum_threshold = best_accuracy - cube_err_table[idx_better_agg[i][j]][i][j]
                            for k in range(len(aggregators)):
                                if k != idx_better_agg[i][j] :
                                    other_acc_confidence = cube_acc_table[k][i][j] + cube_err_table[k][i][j]
                                    if other_acc_confidence >= minimum_threshold:
                                        best_agg_list.append((aggregators[k]["name"] + "_" + str(np.round(cube_acc_table[k][i][j], 4)) + "_" + str(np.round(cube_err_table[k][i][j], 4))))
                                    
                            matrix[i][j] = "\n".join(best_agg_list)

                file_name = str(
                    dataset_name + "_"
                    + model_name + "_"
                    + self.custom_dict_to_str(data_dist["name"]) + "_"
                    + pre_agg_names + "_"
                    + "best_aggregation"
                )
                
                column_names = [str(alpha) for alpha in alpha_list]
                row_names = [str(nb_byzantine) for nb_byzantine in nb_byz]
                column_names.reverse()
                
                pdf_pages = PdfPages(path_to_table + "/" + file_name + ".pdf")

                plt.figure(figsize=(8, 6))
                plt.axis('off')

                plt.table(cellText=matrix,
                        rowLabels=column_names,
                        colLabels=row_names,
                        cellLoc= 'center',
                        loc='center',
                        bbox=[0.1, 0.15, 0.9, 0.9]).scale(1, 1)
                
                #plt.rcParams.update({'font.size': 36})
                

                pdf_pages.savefig()
                pdf_pages.close()

                plt.close()
                    #except Exception as e:
                    #    print(e)
    
    def best_hyperparameters_for_pair_hetero_byz(self, path_to_results, path_to_table, path_to_hyperparameters):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_table):
            os.makedirs(path_to_table)
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if not isinstance(nb_workers, list):
            nb_workers = [nb_workers]
            change = True
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        
        data_folder = data["general"]["data_folder"]
        if data_folder is None:
            data_folder = "./data"

        for data_dist in data_distributions:
            alpha_list = [""]
            if data_dist["name"] == "dirichlet_niid":
                alpha_list = data_dist["parameters"]["alpha"]
            if data_dist["name"] == "gamma_similarity_niid":
                alpha_list = data_dist["parameters"]["gamma"]
            for pre_agg in pre_aggregators:
                pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
                pre_agg_names = "_".join(pre_agg_list_names)
                for agg in aggregators:
                    #try:
                    hyp_table = np.zeros((len(alpha_list), len(nb_byz), 2))
                    for nb_nodes in nb_workers:
                        for y, nb_byzantine in enumerate(nb_byz):
                            if change:
                                nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                            for x, alpha in enumerate(alpha_list):
                                hyperparameters_file_name = str(dataset_name + "_"
                                                                + model_name + "_n_"
                                                                + str(nb_nodes) + "_f_"
                                                                + str(nb_byzantine) + "_"
                                                                + self.custom_dict_to_str(data_dist["name"])
                                                                + str(alpha) + "_"
                                                                + pre_agg_names + "_"
                                                                + agg["name"]
                                                                + ".txt")
                                
                                hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)

                                lr = hyperparameters[0]
                                momentum = hyperparameters[1]

                                hyp_table[len(hyp_table)-1-x][y][0] = lr
                                hyp_table[len(hyp_table)-1-x][y][1] = momentum

                    matrix = [[0 for _ in range(len(alpha_list))] for _ in range(len(nb_byz))]
                    for i in range(len(alpha_list)):
                        for j in range(len(nb_byz)):
                            matrix[i][j] = (hyp_table[i][j][0], hyp_table[i][j][1])

                    file_name = str(
                        dataset_name + "_"
                        + model_name + "_"
                        + self.custom_dict_to_str(data_dist["name"]) + "_"
                        + pre_agg_names + "_"
                        + self.custom_dict_to_str(agg["name"])
                    )
                    
                    column_names = [str(alpha) for alpha in alpha_list]
                    row_names = [str(nb_byzantine) for nb_byzantine in nb_byz]
                    column_names.reverse()
                    
                    pdf_pages = PdfPages(path_to_table + "/" + file_name + ".pdf")

                    plt.figure(figsize=(8, 6))
                    plt.axis('off')

                    plt.table(cellText=matrix,
                            rowLabels=column_names,
                            colLabels=row_names,
                            cellLoc= 'center',
                            loc='center',
                            bbox=[0.15, 0.15, 0.9, 0.9])
                    

                    pdf_pages.savefig()
                    pdf_pages.close()

                    plt.close()
                        #except Exception as e:
                        #    print(e)


table = Table()
#table.best_hyperparameters_for_pair_hetero_byz("./results", "./best_hyperparameters_table", "./best_hyperparameters")
table.best_agg_for_pair_hetero_byz("./results", "./best_aggregations", "./best_hyperparameters")


import math
import byzfl.utils.misc as misc

class NNM(object):

    r"""
    Apply the Nearest Neighbor Mixing pre-aggregation rule [1]_:

    .. math::

        \mathrm{NNM}_{f} \ (x_1, \dots, x_n) = \left(\frac{1}{n-f}\sum_{i \in \mathit{N}_{1}} x_i \ \ , \ \dots \ ,\ \  \frac{1}{n-f}\sum_{i \in \mathit{N}_{n}} x_i \right)
        
    where :math:`\mathit{N}_{i}` is the set of the :math:`n - f` nearest neighbors of :math:`x_i` in :math:`\{x_1, \dots , x_n\}`.

    
    Initialization parameters
    --------------------------
    f : int, optional
        Number of faulty vectors. Set to 0 by default.
    
    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------
        
    >>> import byzfl
    >>> agg = byzfl.NNM(1)

    Using numpy arrays
        
    >>> import numpy as np
    >>> x = np.array([[1., 2., 3.],       # np.ndarray
    >>>               [4., 5., 6.], 
    >>>               [7., 8., 9.]])
    >>> agg(x)
    array([[2.5 3.5 4.5]
            [2.5 3.5 4.5]
            [5.5 6.5 7.5]])

    Using torch tensors
        
    >>> import torch
    >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
    >>>                   [4., 5., 6.], 
    >>>                   [7., 8., 9.]])
    >>> agg(x)
    tensor([[2.5000, 3.5000, 4.5000],
            [2.5000, 3.5000, 4.5000],
            [5.5000, 6.5000, 7.5000]])

    Using list of numpy arrays

    >>> import numppy as np
    >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
    >>>      np.array([4., 5., 6.]), 
    >>>      np.array([7., 8., 9.])]
    >>> agg(x)
    array([[2.5 3.5 4.5]
            [2.5 3.5 4.5]
            [5.5 6.5 7.5]])

    Using list of torch tensors
        
    >>> import torch
    >>> x = [torch.tensor([1., 2., 3.]),  # list of  torch.tensor 
    >>>      torch.tensor([4., 5., 6.]), 
    >>>      torch.tensor([7., 8., 9.])]
    >>> agg(x)
    tensor([[2.5000, 3.5000, 4.5000],
            [2.5000, 3.5000, 4.5000],
            [5.5000, 6.5000, 7.5000]])


    References
    ----------

    .. [1] Allouah, Y., Farhadkhani, S., Guerraoui, R., Gupta, N., Pinot, R.,
           & Stephan, J. (2023, April). Fixing by mixing: A recipe for optimal
           byzantine ml under heterogeneity. In International Conference on 
           Artificial Intelligence and Statistics (pp. 1232-1300). PMLR.    

    """

    def __init__(self, f=0):
        self.f = f

    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        misc.check_type(self.f, int)

        distance = misc.distance_tool(vectors)

        dist = distance.cdist(vectors, vectors)
        k = len(vectors) - self.f
        indices = tools.argpartition(dist, k-1, axis = 1)[:,:k]
        return tools.mean(vectors[indices], axis = 1)

class Bucketing(object):

    r"""
    Apply the Bucketing pre-aggregation rule [1]_:

    .. math::

        \mathrm{Bucketing}_{s} \ (x_1, \dots, x_n) = 
        \left(\frac{1}{s}\sum_{i=0}^s x_{\pi(i)} \ \ , \ \ 
        \frac{1}{s}\sum_{i=s+1}^{2s} x_{\pi(i)} \ \ , \ \dots \ ,\ \  
        \frac{1}{s}\sum_{i=\left(\lceil n/s \rceil-1\right)s+1}^{n} x_{\pi(i)} \right)

    where \\(\\pi\\) is a random permutation of  \\([n]\\).

    Initialization parameters
    --------------------------
    s : int, optional
        Number of vectors per bucket. Set to 1 by default.
    
    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------

    >>> import byzfl
    >>> agg = byzfl.Bucketing(2)

    Using numpy arrays
        
    >>> import numpy as np
    >>> x = np.array([[1., 2., 3.],       # np.ndarray
    >>>               [4., 5., 6.], 
    >>>               [7., 8., 9.]])
    >>> agg(x)
    array([[4. 5. 6.]
            [4. 5. 6.]])

    Using torch tensors
        
    >>> import torch
    >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
    >>>                   [4., 5., 6.], 
    >>>                   [7., 8., 9.]])
    >>> agg(x)
    tensor([[5.5000, 6.5000, 7.5000],
            [1.0000, 2.0000, 3.0000]])

    Using list of numpy arrays

    >>> import numppy as np
    >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
    >>>      np.array([4., 5., 6.]), 
    >>>      np.array([7., 8., 9.])]
    >>> agg(x)
    array([[4. 5. 6.]
            [4. 5. 6.]])

    Using list of torch tensors
        
    >>> import torch
    >>> x = [torch.tensor([1., 2., 3.]),  # list of  torch.tensor 
    >>>      torch.tensor([4., 5., 6.]), 
    >>>      torch.tensor([7., 8., 9.])]
    >>> agg(x)
    tensor([[5.5000, 6.5000, 7.5000],
            [1.0000, 2.0000, 3.0000]])

        
    Note
    ----
        
    The results when using torch tensor and numpy array differ as it 
    depends on random permutation that are not necessary the same


    References
    ----------

    .. [1] Karimireddy, S. P., He, L., & Jaggi, M. (2020). Byzantine-robust 
           learning on heterogeneous datasets via bucketing. International 
           Conference on Learning Representations 2022.
    """

    def __init__(self, s=1):
        self.s = s

    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        misc.check_type(self.s, int)

        random = misc.random_tool(vectors)

        vectors = random.permutation(vectors)
        nb_buckets = int(math.floor(len(vectors) / self.s))
        buckets = vectors[:nb_buckets * self.s]
        buckets = tools.reshape(buckets, (nb_buckets, self.s, len(vectors[0])))
        output = tools.mean(buckets, axis = 1)
        
        # Adding the last incomplete bucket if it exists
        if nb_buckets != len(vectors) / self.s :
            last_mean = tools.mean(vectors[nb_buckets * self.s:], axis = 0)
            last_mean = last_mean.reshape(1,-1)
            output = tools.concatenate((output, last_mean), axis = 0)
        return output


class Identity(object):

    def __init__(self):
        pass

    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        return tools.copy(vectors)


class Clipping(object):

    r"""
    Apply the static Clipping pre-aggregation rule:

    .. math::

        \mathrm{Clipping}_{c} \ (x_1, \dots, x_n) = 
        \left( \min\left\{1, \frac{c}{\|x_1\|}\right\} x_1 \ \ , \ \dots \ ,\ \  
        \min\left\{1, \frac{c}{\|x_n\|}\right\} x_n \right)

    Initialization parameters
    --------------------------
    c : float, optional
        Static clipping threshold. Set to 2 by default.
    
    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------
        
    >>> import byzfl
    >>> agg = byzfl.Clipping(2)

    Using numpy arrays
        
    >>> import numpy as np
    >>> x = np.array([[1., 2., 3.],       # np.ndarray
    >>>               [4., 5., 6.], 
    >>>               [7., 8., 9.]])
    >>> agg(x)
    array([[0.53452248, 1.06904497, 1.60356745],
            [0.91168461, 1.13960576, 1.36752692],
            [1.00514142, 1.14873305, 1.29232469]])

    Using torch tensors
        
    >>> import torch
    >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
    >>>                   [4., 5., 6.], 
    >>>                   [7., 8., 9.]])
    >>> agg(x)
    tensor([[0.5345, 1.0690, 1.6036],
            [0.9117, 1.1396, 1.3675],
            [1.0051, 1.1487, 1.2923]])

    Using list of numpy arrays

    >>> import numppy as np
    >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
    >>>      np.array([4., 5., 6.]), 
    >>>      np.array([7., 8., 9.])]
    >>> agg(x)
    array([[0.53452248, 1.06904497, 1.60356745],
            [0.91168461, 1.13960576, 1.36752692],
            [1.00514142, 1.14873305, 1.29232469]])

    Using list of torch tensors
        
    >>> import torch
    >>> x = [torch.tensor([1., 2., 3.]),  # list of  torch.tensor 
    >>>      torch.tensor([4., 5., 6.]), 
    >>>      torch.tensor([7., 8., 9.])]
    >>> agg(x)
    tensor([[0.5345, 1.0690, 1.6036],
            [0.9117, 1.1396, 1.3675],
            [1.0051, 1.1487, 1.2923]])
 
    """
    
    def __init__(self, c=2):
        self.c = c
    
    def _clip_vector(self, vector):
        tools, vector = misc.check_vectors_type(vector)
        vector_norm = tools.linalg.norm(vector)
        if vector_norm > self.c:
            vector = tools.multiply(vector, self.c / vector_norm)
        return vector

    def __call__(self, vectors):
        for i in range(len(vectors)):
            vectors[i] = self._clip_vector(vectors[i])
        return vectors


class ARC(object):
    
    r"""

    Apply the Adaptive Robust Clipping pre-aggregation rule [1]_:

    .. math::

        \mathrm{ARC}_{f} \ (x_1, \dots, x_n) = 
        \left( \min\left\{1, \frac{x_{\pi(k)}}{\|x_1\|}\right\} x_1 \ \ , \ \dots \ ,\ \  
        \min\left\{1, \frac{x_{\pi(k)}}{\|x_n\|}\right\} x_n \right)

    where :math:`k = \lfloor 2 \frac{f}{n} (n-f)\rfloor` and \\(\\pi\\) is a permutation such that :math:`\|x_{\pi(1)}\|_2 \geq \dots \geq \|x_{\pi(n)}\|_2`.

    Initialization parameters
    --------------------------
    f : int, optional
        Number of faulty vectors. Set to 0 by default.
    
    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------
        
        >>> import byzfl
        >>> agg = byzfl.ARC(1)

        Using numpy arrays
            
        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([[1.        , 2.        , 3.        ],
               [4.        , 5.        , 6.        ],
               [4.41004009, 5.04004582, 5.67005155]])

        Using torch tensors
            
        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([[1.0000, 2.0000, 3.0000],
                [4.0000, 5.0000, 6.0000],
                [4.4100, 5.0400, 5.6701]])
        
        Using list of numpy arrays

        >>> import numppy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([[1.        , 2.        , 3.        ],
               [4.        , 5.        , 6.        ],
               [4.41004009, 5.04004582, 5.67005155]])

        Using list of torch tensors
            
        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of  torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([[1.0000, 2.0000, 3.0000],
                [4.0000, 5.0000, 6.0000],
                [4.4100, 5.0400, 5.6701]])

    References
    ----------

    .. [1] Youssef Allouah, Rachid Guerraoui, Nirupam Gupta, Ahmed Jellouli, Geovani Rizk, and John Stephan. 2024.
           The Vital Role of Gradient Clipping in Byzantine-Resilient Distributed Learning. arXiv:2405.14432 [cs.LG] https://arxiv.org/abs/ 2405.14432
    """

    def __init__(self, f=0):
        self.f = f
    
    def _clip_vector(self, vector, clip_threshold):
        tools, vector = misc.check_vectors_type(vector)
        vector_norm = tools.linalg.norm(vector)

        if vector_norm > clip_threshold:
            vector = tools.multiply(vector, (clip_threshold / vector_norm))
        return vector

    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        magnitudes = [(tools.linalg.norm(vector), vector_id) 
                      for vector_id, vector in enumerate(vectors)]
        magnitudes.sort(key=lambda x:x[0])
        nb_vectors = len(vectors)
        nb_clipped = int((2 * self.f / nb_vectors) * (nb_vectors - self.f))
        cut_off_value = nb_vectors - nb_clipped
        f_largest = magnitudes[cut_off_value:]
        clipping_threshold = magnitudes[cut_off_value - 1][0]
        for _, vector_id in f_largest:
            vectors[vector_id] = self._clip_vector(vectors[vector_id], clipping_threshold)
        return vectors


import inspect

from byzfl.aggregators import aggregators
from byzfl.aggregators import preaggregators

class RobustAggregator(object):

    """
    Description
    -----------
    Class to apply all the pre-aggregations and the aggregation
    to the vectors.

    Parameters
    ----------
    aggregator-info : dict 
        Dictionary with the keys "name" and "parameters" 
        of the aggregation defined.
    pre-agg-info : list
        List of dictionaries (one for every pre_agg function)
        where every dictionary have the keys "name" and "parameters" defined.

    Methods
    -------
    """

    def __init__(self, aggregator_info, pre_agg_list=[]):

        self.aggregator = getattr(aggregators, aggregator_info["name"])

        signature_agg = inspect.signature(self.aggregator.__init__)

        agg_parameters = {}

        for parameter in signature_agg.parameters.values():
            param_name = parameter.name
            if param_name in aggregator_info["parameters"]:
                agg_parameters[param_name] = aggregator_info["parameters"][param_name]
        
        self.aggregator = self.aggregator(**agg_parameters)

        self.pre_agg_list = []

        for pre_agg_info in pre_agg_list:

            pre_agg = getattr(
                preaggregators, 
                pre_agg_info["name"]
            )

            signature_pre_agg = inspect.signature(pre_agg.__init__)

            pre_agg_parameters = {}

            for parameter in signature_pre_agg.parameters.values():
                param_name = parameter.name
                if param_name in pre_agg_info["parameters"]:
                    pre_agg_parameters[param_name] = pre_agg_info["parameters"][param_name]

            pre_agg = pre_agg(**pre_agg_parameters)

            self.pre_agg_list.append(pre_agg)

    def __call__(self, vectors):
        """
        Description
        -----------
        Apply pre-aggregations and aggregations to the vectors

        Parameters
        ----------
        vectors : (list or np.ndarray or torch.Tensor)
            A list of vectors or a matrix (2D array/tensor) where each
            row represents a vector.

        Returns
        -------
        ndarray or torch.Tensor
            Returns a vector with the pre_aggreagtions and the aggregation applied to the input vectors
        """
        for pre_agg in self.pre_agg_list:
            vectors = pre_agg(vectors)

        aggregate_vector = self.aggregator(vectors)  

        return aggregate_vector


# coding: utf-8
###
 # @file   models.py
 # @author John Stephan <john.stephan@epfl.ch>
 #
 # @section LICENSE
 #
 # Copyright © 2023 École Polytechnique Fédérale de Lausanne (EPFL).
 # All rights reserved.
 #
 # @section DESCRIPTION
 #
 # Collection of models.
###

import torch
import torch.nn as nn
import torch.nn.functional as F

# ---------------------------------------------------------------------------- #
# Simple fully-connected model, for MNIST
class fc_mnist(nn.Module):
	""" Simple, small fully connected model."""

	def __init__(self):
		""" Model parameter constructor. """
		super().__init__()
		# Build parameters
		self._f1 = nn.Linear(28 * 28, 100)
		self._f2 = nn.Linear(100, 10)

	def forward(self, x):
		""" Model's forward pass. """
		x = F.relu(self._f1(x.view(-1, 28 * 28)))
		x = F.log_softmax(self._f2(x), dim=1)
		return x

# ---------------------------------------------------------------------------- #
# Simple convolutional model, for MNIST
class cnn_mnist(nn.Module):
	""" Simple, small convolutional model."""

	def __init__(self):
		""" Model parameter constructor. """
		super().__init__()
		# Build parameters
		self._c1 = nn.Conv2d(1, 20, 5, 1)
		self._c2 = nn.Conv2d(20, 50, 5, 1)
		self._f1 = nn.Linear(800, 500)
		self._f2 = nn.Linear(500, 10)

	def forward(self, x):
		""" Model's forward pass. """
		x = F.relu(self._c1(x))
		x = F.max_pool2d(x, 2, 2)
		x = F.relu(self._c2(x))
		x = F.max_pool2d(x, 2, 2)
		x = F.relu(self._f1(x.view(-1, 800)))
		x = F.log_softmax(self._f2(x), dim=1)
		return x

# ---------------------------------------------------------------------------- #
# Simple logistic regression model for MNIST
class logreg_mnist(nn.Module):
	""" Simple logistic regression model."""

	def __init__(self):
		""" Model parameter constructor. """
		super().__init__()
		# Build parameters
		self._linear = nn.Linear(784, 10)

	def forward(self, x):
		""" Model's forward pass. """
		return torch.sigmoid(self._linear(x.view(-1, 784)))

# ---------------------------------------------------------------------------- #
#JS: Simple logistic regression model (for phishing)
class logreg_phishing(nn.Module):
	""" Simple logistic regression model."""

	def __init__(self, din, dout=1):
		""" Model parameter constructor.
		Args:
			din  Number of input dimensions
			dout Number of output dimensions
		"""
		super().__init__()
		# Store model parameters
		self._din  = din
		self._dout = dout
		# Build parameters
		self._linear = nn.Linear(din, dout)

	def forward(self, x):
		""" Model's forward pass. """
		return torch.sigmoid(self._linear(x.view(-1, self._din)))


# ---------------------------------------------------------------------------- #
#JS: Simple convolutional model, for CIFAR-10/100 (3 input channels)
class cnn_cifar_old(nn.Module):
  """ Simple, small convolutional model."""

  def __init__(self):
    """ Model parameter constructor."""
    super().__init__()
    # Build parameters
    self._c1 = nn.Conv2d(3, 64, kernel_size=3, padding=1)
    self._b1 = nn.BatchNorm2d(self._c1.out_channels)
    self._c2 = nn.Conv2d(self._c1.out_channels, 64, kernel_size=3, padding=1)
    self._b2 = nn.BatchNorm2d(self._c2.out_channels)
    self._m1 = nn.MaxPool2d(2)
    self._d1 = nn.Dropout(p=0.25)
    self._c3 = nn.Conv2d(self._c2.out_channels, 128, kernel_size=3, padding=1)
    self._b3 = nn.BatchNorm2d(self._c3.out_channels)
    self._c4 = nn.Conv2d(self._c3.out_channels, 128, kernel_size=3, padding=1)
    self._b4 = nn.BatchNorm2d(self._c4.out_channels)
    self._m2 = nn.MaxPool2d(2)
    self._d2 = nn.Dropout(p=0.25)
    self._d3 = nn.Dropout(p=0.25)
    self._f1 = nn.Linear(8192, 128)
    self._f2 = nn.Linear(self._f1.out_features, 10)

  def forward(self, x):
    """ Model's forward pass. """
    activation = F.relu
    flatten    = lambda x: x.view(x.shape[0], -1)
    logsoftmax = F.log_softmax
    # Forward pass
    x = self._c1(x)
    x = activation(x)
    x = self._b1(x)
    x = self._c2(x)
    x = activation(x)
    x = self._b2(x)
    x = self._m1(x)
    x = self._d1(x)
    x = self._c3(x)
    x = activation(x)
    x = self._b3(x)
    x = self._c4(x)
    x = activation(x)
    x = self._b4(x)
    x = self._m2(x)
    x = self._d2(x)
    x = flatten(x)
    x = self._f1(x)
    x = activation(x)
    x = self._d3(x)
    x = self._f2(x)
    x = logsoftmax(x, dim=1)
    return x

class FedBatchNorm1d(nn.Module):
    def __init__(self, 
                 num_features, 
                 num_workers,
                 eps=1e-05, 
                 momentum=0.1, 
                 affine=True, 
                 track_running_stats=True, 
                 device=None, 
                 dtype=None):
        super(FedBatchNorm1d, self).__init__()
        self.momentum = momentum
        self.eps = eps
        self.affine = affine
        self.track_running_stats = track_running_stats
        self.device = device
        self.dtype = dtype
        self.register_buffer('previous_batch_var', torch.zeros(num_features).to(device))
        self.register_buffer('previous_batch_mean', torch.zeros(num_features).to(device))
        if self.affine == True:
            self.weight = nn.Parameter(torch.ones(num_features).to(device))
            self.bias = nn.Parameter(torch.zeros(num_features).to(device))
        self.running_mean = torch.zeros(num_features).to(device)
        self.running_var = torch.ones(num_features).to(device)
        self.num_batches_tracked = torch.zeros(1).to(device)
        self.num_workers = num_workers
        self.b = nn.BatchNorm2d(2)

    def forward(self, input):
        if self.num_batches_tracked > 0:
            self.running_mean  = (1-self.momentum) * self.running_mean + self.momentum * self.previous_batch_mean
            nb_data = self.num_workers*len(input) - 1
            constructed_var = (self.num_workers/nb_data) * self.previous_batch_var - ((nb_data+1)/nb_data) * (self.running_mean**2) 
            self.running_var = (1-self.momentum) * self.running_var + self.momentum * constructed_var
        
        self.previous_batch_mean = input.mean(dim = 0)
        self.previous_batch_var = (input**2).sum(dim = 0)
        
        output = ((input - self.running_mean)/torch.sqrt(self.running_var + self.eps))
        self.num_batches_tracked = self.num_batches_tracked + 1
        if self.affine == False:
            return output
        return output * self.weight + self.bias


class FedBatchNorm(nn.BatchNorm2d):
    def __init__(self, num_features, nb_clients, eps=1e-5, momentum=0.1,
                 affine=True, track_running_stats=True):
        super(FedBatchNorm, self).__init__(
            num_features, eps, momentum, affine, track_running_stats)
        self.n = nb_clients

    def forward(self, input):
        self._check_input_dim(input)

        mean = self.running_mean.clone()
        var = self.running_var.clone()
        
        if self.training:
            
            current_mean = input.mean([0, 2, 3])
            current_var = input.var([0, 2, 3], correction=0)
      
            K = input.numel() / self.num_features
            Kn = K * self.n
            with torch.no_grad():
                self.running_mean = self.momentum * current_mean + (1 - self.momentum) * self.running_mean
                self.running_var = self.momentum * current_var * (Kn / (Kn - 1)) + (1 - self.momentum) * self.running_var

        input = (input - mean[None, :, None, None]) / (torch.sqrt(var[None, :, None, None] + self.eps))
        if self.affine:
            input = input * self.weight[None, :, None, None] + self.bias[None, :, None, None]

        return input


class cnn_cifar_old_with_fbn(nn.Module):
  """ Simple, small convolutional model."""

  def __init__(self, num_clients):
    """ Model parameter constructor."""
    super().__init__()
    # Build parameters
    self._c1 = nn.Conv2d(3, 64, kernel_size=3, padding=1)
    self._b1 = FedBatchNorm(self._c1.out_channels, num_clients)
    self._c2 = nn.Conv2d(self._c1.out_channels, 64, kernel_size=3, padding=1)
    self._b2 = FedBatchNorm(self._c2.out_channels, num_clients)
    self._m1 = nn.MaxPool2d(2)
    self._d1 = nn.Dropout(p=0.25)
    self._c3 = nn.Conv2d(self._c2.out_channels, 128, kernel_size=3, padding=1)
    self._b3 = FedBatchNorm(self._c3.out_channels, num_clients)
    self._c4 = nn.Conv2d(self._c3.out_channels, 128, kernel_size=3, padding=1)
    self._b4 = FedBatchNorm(self._c4.out_channels, num_clients)
    self._m2 = nn.MaxPool2d(2)
    self._d2 = nn.Dropout(p=0.25)
    self._d3 = nn.Dropout(p=0.25)
    self._f1 = nn.Linear(8192, 128)
    self._f2 = nn.Linear(self._f1.out_features, 10)

  def forward(self, x):
    """ Model's forward pass. """
    activation = F.relu
    flatten    = lambda x: x.view(x.shape[0], -1)
    logsoftmax = F.log_softmax
    # Forward pass
    x = self._c1(x)
    x = activation(x)
    x = self._b1(x)
    x = self._c2(x)
    x = activation(x)
    x = self._b2(x)
    x = self._m1(x)
    x = self._d1(x)
    x = self._c3(x)
    x = activation(x)
    x = self._b3(x)
    x = self._c4(x)
    x = activation(x)
    x = self._b4(x)
    x = self._m2(x)
    x = self._d2(x)
    x = flatten(x)
    x = self._f1(x)
    x = activation(x)
    x = self._d3(x)
    x = self._f2(x)
    x = logsoftmax(x, dim=1)
    return x

class cifar_Net(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 20, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(20, 200, 5)
        self.fc1 = nn.Linear(200 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        return self.fc3(x)

class cnn_cifar(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(3, 20, 5, padding=2)
        self.conv2 = nn.Conv2d(self.conv1.out_channels, 100, 5, padding=2)
        self.conv3 = nn.Conv2d(self.conv2.out_channels, 200, 5, padding=2)
        self.pool = nn.MaxPool2d(2, 2)
        self.fc1 = nn.Linear(self.conv3.out_channels * 4 * 4, 512)
        self.fc2 = nn.Linear(self.fc1.out_features, 256)
        self.fc3 = nn.Linear(self.fc2.out_features, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = self.pool(F.relu(self.conv3(x)))
        x = torch.flatten(x, 1)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        return self.fc3(x)


#################################################################################################
#JS: Resnet models

class BasicBlock(nn.Module):
    expansion = 1
 
    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
 
        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )
 
    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out
 
 
class Bottleneck(nn.Module):
    expansion = 4
 
    def __init__(self, in_planes, planes, stride=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion*planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion*planes)
 
        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )
 
    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out
 
 
class ResNet(nn.Module):
    def __init__(self, block, num_blocks, num_classes=10):
        super(ResNet, self).__init__()
        self.in_planes = 64
 
        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2)
        self.linear = nn.Linear(512*block.expansion, num_classes)
 
    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)
 
    def forward(self, x, out_feature=False):
        x = self.conv1(x)

        x = self.bn1(x)
        out = F.relu(x)

        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = F.avg_pool2d(out, 4)
        feature = out.view(out.size(0), -1)
        out = self.linear(feature)
        if out_feature == False:
            return out
        else:
            return out,feature
 
 
def ResNet18(num_classes=10):
    return ResNet(BasicBlock, [2,2,2,2], num_classes)
 
def ResNet34(num_classes=10):
    return ResNet(BasicBlock, [3,4,6,3], num_classes)
 
def ResNet50(num_classes=10):
    return ResNet(Bottleneck, [3,4,6,3], num_classes)
 
def ResNet101(num_classes=10):
    return ResNet(Bottleneck, [3,4,23,3], num_classes)
 
def ResNet152(num_classes=10):
    return ResNet(Bottleneck, [3,8,36,3], num_classes)


import collections

import torch

import byzfl.pipeline.models as models
from byzfl.utils.conversion import flatten_dict, unflatten_dict, unflatten_generator

class ModelBaseInterface(object):
    """
    Description
    -----------
    This class serves as an abstract interface that defines the methods 
    required for classes that encapsulate a model. All subclasses that 
    contain a model should inherit from this class to ensure they implement 
    the necessary methods for handling model-related operations and information 
    exchange.

    Parameters
    ----------
        All this parameters should be passed in a dictionary that contains the following keys.
    model-name : str 
        Indicates the model to be used
    device : str
        Name of the device used
    learning-rate : float 
        Learning rate
    weight-decay : float 
        Regularization used
    milestones : list 
        List of the milestones, where the learning rate decay should be applied
    learning-rate-decay : float 
        Rate decreases over time during training

    Methods
    --------
    """
    def __init__(self, params):
        model_name = params["model_name"]

        self.device = params["device"]

        if "fbn" in model_name:
            self.model = torch.nn.DataParallel(getattr(models, model_name)(params["nb_workers"])).to(self.device)
            self.use_fbn = True
        else:
            self.model = torch.nn.DataParallel(getattr(models, model_name)()).to(self.device)

        self.optimizer = torch.optim.SGD(
            self.model.parameters(), 
            lr = params["learning_rate"], 
            weight_decay = params["weight_decay"]
        )

        self.scheduler = torch.optim.lr_scheduler.MultiStepLR(
            self.optimizer,
            milestones = params["milestones"],
            gamma = params["learning_rate_decay"]
        )

        self.batch_norm_keys = []
        self.running_mean_key_list = []
        self.running_var_key_list = []
    
    def get_flat_parameters(self):
        """
        Description
        -----------
        Get the gradients of the model in a flat array

        Returns
        -------
        List of the gradients
        """
        return flatten_dict(self.model.state_dict())
    
    def get_flat_gradients(self):
        """
        Description
        -----------
        Get the gradients of the model in a flat array

        Returns
        -------
        List of the gradients
        """
        return flatten_dict(self.get_dict_gradients())
    
    def get_dict_parameters(self):
        """
        Description
        ------------
        Get the gradients of the model in a dictionary.

        Returns
        -------
        Dicctionary where the keys are the name of the parameters
        and de values are the gradients.
        """
        return self.model.state_dict()

    def get_dict_gradients(self):
        """
        Description
        ------------
        Get the gradients of the model in a dictionary.

        Returns
        -------
        Dicctionary where the keys are the name of the parameters
        and the values are the gradients.
        """
        new_dict = collections.OrderedDict()
        for key, value in self.model.named_parameters():
            new_dict[key] = value.grad
        return new_dict
    
    def get_batch_norm_stats(self):
        """
        Description
        ------------
        Get the batch norm stats of the model in a dictionary.

        Returns
        -------
        Dicctionary where the keys are the name of the parameters
        of batch norm stats and their values.
        """
        running_mean_stats = collections.OrderedDict()
        state_dict = self.model.state_dict()
        for key in self.running_mean_key_list:
            running_mean_stats[key] = state_dict[key].clone()

        running_var_stats = collections.OrderedDict()
        for key in self.running_var_key_list:
            running_var_stats[key] = state_dict[key].clone()
        return running_mean_stats, running_var_stats

    def get_flat_batch_norm_stats(self):
        """
        Description
        ------------
        Get the batch norm stats of the model in a flatten array.

        Returns
        -------
        Array with the values of the batch norm stats.
        """
        running_mean_stats, running_var_stats = self.get_batch_norm_stats()
        return flatten_dict(running_mean_stats), flatten_dict(running_var_stats)
    
    def set_parameters(self, flat_vector):
        """
        Description
        -----------
        Sets the model parameters given a flat vector.

        Parameters
        ----------
        flat_vector : list 
            Flat list with the parameters
        """
        new_dict = unflatten_dict(self.model.state_dict(), flat_vector)
        self.model.load_state_dict(new_dict)

    def set_gradients(self, flat_vector):
        """
        Description
        -----------
        Sets the model gradients given a flat vector.

        Parameters
        ----------
        flat_vector : list
            Flat list with the parameters
        """
        new_dict = unflatten_generator(self.model.named_parameters(), flat_vector)
        for key, value in self.model.named_parameters():
            value.grad = new_dict[key].clone().detach()
    
    def set_model_state(self, state_dict):
        """
        Description
        -----------
        Sets the state_dict of the model for the state_dict given by parameter.

        Parameters
        ----------
        state_dict : dict 
            State_dict from a model
        """
        self.model.load_state_dict(state_dict)

    def compute_batch_norm_keys(self):
        """
        Description
        -----------
        Compute batch normalization keys.

        """
        for key in self.model.state_dict().keys():
            if "running_mean" in key:
                self.running_mean_key_list.append(key)
                self.batch_norm_keys.append(key.split(".")[0])
            elif "running_var" in key:
                self.running_var_key_list.append(key)            
    
    def use_batch_norm(self):
        """
        Description
        -----------
        Getter to determine whether the model is using Batch Normalization.

        Returns
        -------
        bool
            A boolean indicating whether the model is utilizing Batch Normalization.
        """
        return len(self.batch_norm_keys) > 0


    def use_federated_batch_norm(self):
        """
        Description
        -----------
        Getter to determine whether the model is using Federated Batch Normalization.

        Returns
        -------
        bool
            A boolean indicating whether the model is utilizing Federated Batch Normalization.
        """
        return self.use_fbn



import byzfl.attacks.attacks as attacks
import byzfl.attacks.attack_optimizers as attack_optimizers

class ByzantineWorker():
    """
    Description
    -----------
    Class that is responsible for carrying out the byzantine attacks.

    Parameters
    ----------
        All this parameters should be passed in a dictionary that contains the following keys.
    nb_byz : int
        Number of byzantine nodes that has the byzantine worker.
    attack_name : str
        Name of the attack
    attack_parameters : dict
        Dictionary with the parameters of the attack 
        where every key is the name of the paramater and their value is 
        the value of the parameter.
    attack_optimizer_name : (str, optional)
        Name of the optimizer to be used to find the best attack parameters.
    attack_optimizer_parameters : (dict, optional)
        Dictionary with the parameters of the optimizer where every key is 
        the name of the paramater and their value is the value of the parameter.
    aggregator_info : dict
        Dictionary with the keys "name" and "parameters" defined for the aggregator.
    pre_agg_list : list 
        List of dictionaries (one for every pre_agg function)
        where every dictionary have the keys "name" and "parameters" defined.
    

    Methods
    -------
    """
    def __init__(self, params):
        self.nb_real_byz = params["nb_byz"]
        self.attack = getattr(
            attacks, 
            params["attack_name"]
        )(**params["attack_parameters"])

        self.optimizer = None
        self.optimizer_batch_norm = None
        
        if params["attack_optimizer_name"] is not None:
            optimizer_params = {
                "agg_info": params["aggregator_info"],
                "pre_agg_list": params["pre_agg_list"],
                "nb_byz": params["nb_byz"],
                **params["attack_optimizer_parameters"]
            }
            self.optimizer = getattr(
                attack_optimizers, 
                params["attack_optimizer_name"]
            )(**optimizer_params)

            self.optimizer_batch_norm = getattr(
                attack_optimizers, 
                params["attack_optimizer_name"]
            )(**optimizer_params)
    
    def apply_attack(self, honest_vectors):
        """
        Computes the byzantine vector and (optimized or not depending
        if the optimizer is configured) and then it returns a list
        with this vector n times, where n = number of byzantine nodes

        Parameters
        ----------
        vectors : list or np.ndarray or torch.Tensor
            A list of vectors or a matrix (2D array/tensor) where each row represents a vector.
        
        Returns
        -------
        Returns a list with the byzantine vector n times, 
        where n = number of byzantine nodes
        """
        if self.nb_real_byz == 0:
            return list()
        if self.optimizer is not None:
            self.optimizer(self.attack, honest_vectors)
        byz_vector = self.attack(honest_vectors)
        
        return [byz_vector] * self.nb_real_byz
    
    def apply_batch_norm_attack(self, honest_vectors):
        """
        Computes the byzantine vector of the batch norms 
        and (optimized or not depending if the optimizer is configured) 
        and then it returns a list with this vector n times, where n = number 
        of byzantine nodes

        Parameters
        ----------
        vectors : list or np.ndarray or torch.Tensor
            A list of vectors or a matrix (2D array/tensor) where each row represents a vector.
        
        Returns
        -------
        Returns a list with the byzantine vector n times, 
        where n = number of byzantine nodes
        """
        if self.nb_real_byz == 0:
            return list()
        if self.optimizer_batch_norm is not None:
            self.optimizer_batch_norm(self.attack, honest_vectors)
        byz_vector = self.attack(honest_vectors)
        
        return [byz_vector] * self.nb_real_byz



import math
import json
import os

import numpy as np
from numpy import genfromtxt
import matplotlib.pyplot as plt
import seaborn as sns

def custom_dict_to_str(d):
        if not d:
            return ''
        else:
            return str(d)

def find_best_hyperparameters(path_to_results, path_hyperparameters):
    try:
        with open(path_to_results+'/settings.json', 'r') as file:
            data = json.load(file)
    except Exception as e:
        print("ERROR: "+ str(e))

    training_seed = data["general"]["training_seed"]
    nb_training_seeds = data["general"]["nb_training_seeds"]
    nb_workers = data["general"]["nb_workers"]
    nb_byz = data["general"]["nb_byz"]
    nb_steps = data["general"]["nb_steps"] #Fixed
    evaluation_delta = data["general"]["evaluation_delta"] #Fixed

    data_distribution_seed = data["model"]["data_distribution_seed"]
    nb_data_distribution_seeds = data["model"]["nb_data_distribution_seeds"]

    model_name = data["model"]["name"] #Fixed
    dataset_name = data["model"]["dataset_name"] #Fixed
    data_distributions = data["model"]["data_distribution"]
    aggregators = data["aggregator"]
    pre_aggregators = data["pre_aggregators"]

    lr_list = data["honest_nodes"]["learning_rate"]
    momentum_list = data["honest_nodes"]["momentum"]
    wd_list = data["honest_nodes"]["weight_decay"]

    attacks = data["attack"]
    
    if isinstance(nb_workers, int):
        nb_workers = [nb_workers]
    if not isinstance(nb_workers, list):
        nb_workers = [nb_workers]
        change = True
    if isinstance(nb_byz, int):
        nb_byz = [nb_byz]
    if isinstance(data_distributions, dict):
        data_distributions = [data_distributions]
    if isinstance(aggregators, dict):
        aggregators = [aggregators]
    if len(pre_aggregators) == 0 or isinstance(pre_aggregators[0], dict):
        pre_aggregators = [pre_aggregators]
    if isinstance(attacks, dict):
        attacks = [attacks]
    if isinstance(lr_list, float):
        lr_list = [lr_list]
    if isinstance(momentum_list, float):
        momentum_list = [momentum_list]
    if isinstance(wd_list, float):
        wd_list = [wd_list]

    nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))

    for nb_nodes in nb_workers:
        for nb_byzantine in nb_byz:
            if change:
                nb_nodes = data["general"]["nb_honest"] + nb_byzantine
            for data_dist in data_distributions:
                distribution_parameter_list = data_dist["distribution_parameter"]
                for distribution_parameter in distribution_parameter_list:
                    for pre_agg in pre_aggregators:
                        pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
                        pre_agg_names = "_".join(pre_agg_list_names)
                        real_hyper_parameters = np.zeros((len(aggregators),3))
                        real_steps = np.zeros((len(aggregators), len(attacks)))
                        for k, agg in enumerate(aggregators):
                            max_acc_config = np.zeros((len(lr_list)*len(momentum_list)*len(wd_list), len(attacks)))
                            hyper_parameters = np.zeros((len(lr_list)*len(momentum_list)*len(wd_list), 3))
                            steps_max_reached = np.zeros((len(lr_list)*len(momentum_list)*len(wd_list), len(attacks)))
                            z = 0
                            for lr in lr_list:
                                for momentum in momentum_list:
                                    for wd in wd_list:
                                        tab_acc = np.zeros((
                                            len(attacks),
                                            nb_data_distribution_seeds,
                                            nb_training_seeds,
                                            nb_accuracies
                                        ))
                                        for i, attack in enumerate(attacks):
                                            for run_dd in range(nb_data_distribution_seeds):
                                                for run in range(nb_training_seeds):
                                                    common_name = str(dataset_name + "_" 
                                                                    + model_name + "_n_" 
                                                                    + str(nb_nodes) + "_f_" 
                                                                    + str(nb_byzantine) + "_d_"
                                                                    + str(nb_byzantine) + "_" 
                                                                    + custom_dict_to_str(data_dist["name"])  + "_"
                                                                    + str(distribution_parameter) + "_" 
                                                                    + custom_dict_to_str(agg["name"]) + "_" 
                                                                    + pre_agg_names + "_" 
                                                                    + custom_dict_to_str(attack["name"]) + "_lr_"
                                                                    + str(lr) + "_mom_"
                                                                    + str(momentum) + "_wd_"
                                                                    + str(wd))
                                                    tab_acc[i][run_dd][run] = genfromtxt(path_to_results+"/"+common_name+"/validation_accuracy_tr_seed_"+str(run+training_seed)+"_dd_seed_" + str(run_dd+data_distribution_seed) + ".txt", delimiter=',')

                                        for i, attack in enumerate(attacks):
                                            avg_accuracy = np.mean(tab_acc[i], axis=1)
                                            avg_accuracy_dd = np.mean(avg_accuracy, axis=0)
                                            idx_max = np.argmax(avg_accuracy_dd)
                                            max_acc_config[z][i] = avg_accuracy_dd[idx_max]
                                            steps_max_reached[z][i] = idx_max * evaluation_delta

                                        hyper_parameters[z][0] = lr
                                        hyper_parameters[z][1] = momentum
                                        hyper_parameters[z][2] = wd
                                        z += 1

                            if not os.path.exists(path_hyperparameters):
                                try:
                                    os.makedirs(path_hyperparameters)
                                except OSError as error:
                                    print(f"Error creating directory: {error}")

                            max_minimum_index = -1
                            maximum_minimum_value = -1
                            for i in range(z):
                                actual_min = np.min(max_acc_config[i])
                                if actual_min > maximum_minimum_value:
                                    max_minimum_index = i
                                    maximum_minimum_value = actual_min

                            real_hyper_parameters[k] = hyper_parameters[max_minimum_index]
                            real_steps[k] = steps_max_reached[max_minimum_index]

                        hyper_parameters_folder = path_hyperparameters + "/hyperparameters"
                        steps_folder = path_hyperparameters + "/better_step"

                        if not os.path.exists(hyper_parameters_folder):
                            os.makedirs(hyper_parameters_folder)

                        if not os.path.exists(steps_folder):
                            os.makedirs(steps_folder)

                        for i, agg in enumerate(aggregators):
                            file_name_hyperparameters = str(dataset_name + "_"
                                                            + model_name + "_n_"
                                                            + str(nb_nodes) + "_f_"
                                                            + str(nb_byzantine) + "_d_"
                                                            + str(nb_byzantine) + "_"
                                                            + custom_dict_to_str(data_dist["name"]) + "_"
                                                            + str(distribution_parameter) + "_"
                                                            + pre_agg_names + "_"
                                                            + agg["name"]
                                                            + ".txt")
                            np.savetxt(hyper_parameters_folder+"/"+file_name_hyperparameters, real_hyper_parameters[i])
                            
                            for j, attack in enumerate(attacks):
                                file_name_steps = str(dataset_name + "_"
                                                    + model_name + "_n_"
                                                    + str(nb_nodes) + "_f_"
                                                    + str(nb_byzantine) + "_d_"
                                                    + str(nb_byzantine) + "_"
                                                    + custom_dict_to_str(data_dist["name"]) + "_"
                                                    + str(distribution_parameter) + "_"
                                                    + pre_agg_names + "_"
                                                    + agg["name"] + "_"
                                                    + custom_dict_to_str(attack["name"])
                                                    + ".txt")
                                np.savetxt(steps_folder+"/"+file_name_steps, np.array([real_steps[i][j]]))

def heat_map_test_accuracy(path_to_results, path_to_hyperparameters, path_to_plot):
    try:
        with open(path_to_results+'/settings.json', 'r') as file:
            data = json.load(file)
    except Exception as e:
        print("ERROR: "+ str(e))

    if not os.path.exists(path_to_plot):
        os.makedirs(path_to_plot)
    
    training_seed = data["general"]["training_seed"]
    nb_training_seeds = data["general"]["nb_training_seeds"]
    nb_workers = data["general"]["nb_workers"]
    nb_byz = data["general"]["nb_byz"]
    nb_steps = data["general"]["nb_steps"] #Fixed
    evaluation_delta = data["general"]["evaluation_delta"] #Fixed

    data_distribution_seed = data["model"]["data_distribution_seed"]
    nb_data_distribution_seeds = data["model"]["nb_data_distribution_seeds"]

    model_name = data["model"]["name"] #Fixed
    dataset_name = data["model"]["dataset_name"] #Fixed
    data_distributions = data["model"]["data_distribution"]
    aggregators = data["aggregator"]
    pre_aggregators = data["pre_aggregators"]

    attacks = data["attack"]
    
    if isinstance(nb_workers, int):
        nb_workers = [nb_workers]
    if not isinstance(nb_workers, list):
        nb_workers = [nb_workers]
        change = True
    if isinstance(nb_byz, int):
        nb_byz = [nb_byz]
    if isinstance(data_distributions, dict):
        data_distributions = [data_distributions]
    if isinstance(aggregators, dict):
        aggregators = [aggregators]
    if len(pre_aggregators) == 0 or isinstance(pre_aggregators[0], dict):
        pre_aggregators = [pre_aggregators]
    if isinstance(attacks, dict):
        attacks = [attacks]

    for pre_agg in pre_aggregators:
        pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
        pre_agg_names = "_".join(pre_agg_list_names)
        for agg in aggregators:
            for data_dist in data_distributions:
                distribution_parameter_list = data_dist["distribution_parameter"]
                heat_map_table = np.zeros((len(distribution_parameter_list), len(nb_byz)))
                for nb_nodes in nb_workers:
                    for y, nb_byzantine in enumerate(nb_byz):
                        if change:
                            nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                        for x, dist_param in enumerate(distribution_parameter_list):
                            hyperparameters_file_name = str(dataset_name + "_"
                                                            + model_name + "_n_"
                                                            + str(nb_nodes) + "_f_"
                                                            + str(nb_byzantine) + "_d_"
                                                            + str(nb_byzantine) + "_"
                                                            + custom_dict_to_str(data_dist["name"]) + "_"
                                                            + str(dist_param) + "_"
                                                            + pre_agg_names + "_"
                                                            + agg["name"]
                                                            + ".txt")
                            
                            hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)

                            lr = hyperparameters[0]
                            momentum = hyperparameters[1]
                            wd = hyperparameters[2]
                            
                            worst_accuracy = np.inf
                            for attack in attacks:
                                test_accuracy_file_name = str(
                                            dataset_name + "_" 
                                            + model_name + "_n_" 
                                            + str(nb_nodes) + "_f_" 
                                            + str(nb_byzantine) + "_d_"
                                            + str(nb_byzantine) + "_"
                                            + custom_dict_to_str(data_dist["name"]) + "_"
                                            + str(dist_param) + "_" 
                                            + custom_dict_to_str(agg["name"]) + "_"
                                            + pre_agg_names + "_"
                                            + custom_dict_to_str(attack["name"]) + "_" 
                                            + "lr_" + str(lr) + "_" 
                                            + "mom_" + str(momentum) + "_" 
                                            + "wd_" + str(wd)
                                )
                                try:
                                    with open(path_to_results+ "/" + test_accuracy_file_name +'/settings.json', 'r') as file:
                                        data = json.load(file)
                                except Exception as e:
                                    print("ERROR: "+ str(e))
                                nb_steps = data["general"]["nb_steps"]
                                nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
                                tab_acc = np.zeros((
                                        nb_data_distribution_seeds,
                                        nb_training_seeds, 
                                        nb_accuracies
                                ))
                                for run_dd in range(nb_data_distribution_seeds):
                                    for run in range(nb_training_seeds):
                                        tab_acc[run_dd][run] = genfromtxt(path_to_results + "/" + test_accuracy_file_name + "/test_accuracy_tr_seed_"+str(run+training_seed)+ "_dd_seed_" + str(run_dd + data_distribution_seed) + ".txt", delimiter=',')
                                tab_acc = tab_acc.mean(axis=1)
                                tab_acc = tab_acc.mean(axis=0)
                                accuracy = np.max(tab_acc)
                                if accuracy < worst_accuracy:
                                    worst_accuracy = accuracy
                                
                            heat_map_table[len(heat_map_table)-1-x][y] = worst_accuracy
                
                file_name = str(
                    "test_"
                    + dataset_name + "_"
                    + model_name + "_"
                    + custom_dict_to_str(data_dist["name"]) + "_"
                    + pre_agg_names + "_"
                    + agg["name"] + ".pdf"
                )
                
                column_names = [str(dist_param) for dist_param in distribution_parameter_list]
                row_names = [str(nb_byzantine) for nb_byzantine in nb_byz]
                column_names.reverse()

                sns.heatmap(heat_map_table, xticklabels=row_names, yticklabels=column_names, annot=True)
                plt.savefig(path_to_plot +"/"+ file_name)
                plt.close()

def best_heat_map_test_accuracy(path_to_results, path_to_hyperparameters, path_to_plot):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_plot):
            os.makedirs(path_to_plot)
        
        training_seed = data["general"]["training_seed"]
        nb_training_seeds = data["general"]["nb_training_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        data_distribution_seed = data["model"]["data_distribution_seed"]
        nb_data_distribution_seeds = data["model"]["nb_data_distribution_seeds"]

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        attacks = data["attack"]

        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if not isinstance(nb_workers, list):
            nb_workers = [nb_workers]
            change = True
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if len(pre_aggregators) == 0 or isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        
        for pre_agg in pre_aggregators:
            pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
            pre_agg_names = "_".join(pre_agg_list_names)
            heat_map_cube = np.zeros((len(aggregators), len(data_distributions[0]["distribution_parameter"]), len(nb_byz)))
            for z, agg in enumerate(aggregators):
                for data_dist in data_distributions:
                    distribution_parameter_list = data_dist["distribution_parameter"]
                    heat_map_table = np.zeros((len(distribution_parameter_list), len(nb_byz)))
                    for nb_nodes in nb_workers:
                        for y, nb_byzantine in enumerate(nb_byz):
                            if change:
                                nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                            for x, dist_param in enumerate(distribution_parameter_list):
                                hyperparameters_file_name = str(dataset_name + "_"
                                                                + model_name + "_n_"
                                                                + str(nb_nodes) + "_f_"
                                                                + str(nb_byzantine) + "_d_"
                                                                + str(nb_byzantine) + "_"
                                                                + custom_dict_to_str(data_dist["name"]) + "_"
                                                                + str(dist_param) + "_"
                                                                + pre_agg_names + "_"
                                                                + agg["name"]
                                                                + ".txt")
                                
                                hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)
                                lr = hyperparameters[0]
                                momentum = hyperparameters[1]
                                wd = hyperparameters[2]
                                
                                worst_accuracy = np.inf
                                for attack in attacks:
                                    test_accuracy_file_name = str(
                                                dataset_name + "_" 
                                                + model_name + "_n_" 
                                                + str(nb_nodes) + "_f_" 
                                                + str(nb_byzantine) + "_d_"
                                                + str(nb_byzantine) + "_"
                                                + custom_dict_to_str(data_dist["name"]) + "_"
                                                + str(dist_param) + "_" 
                                                + custom_dict_to_str(agg["name"]) + "_"
                                                + pre_agg_names + "_"
                                                + custom_dict_to_str(attack["name"]) + "_" 
                                                + "lr_" + str(lr) + "_" 
                                                + "mom_" + str(momentum) + "_" 
                                                + "wd_" + str(wd)
                                    )
                                    try:
                                        with open(path_to_results+ "/" + test_accuracy_file_name +'/settings.json', 'r') as file:
                                            data = json.load(file)
                                    except Exception as e:
                                        print("ERROR: "+ str(e))
                                    nb_steps = data["general"]["nb_steps"]
                                    nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
                                    tab_acc = np.zeros((
                                            nb_data_distribution_seeds,
                                            nb_training_seeds, 
                                            nb_accuracies
                                    ))
                                    for run_dd in range(nb_data_distribution_seeds):
                                        for run in range(nb_training_seeds):
                                            tab_acc[run_dd][run] = genfromtxt(path_to_results + "/" + test_accuracy_file_name + "/test_accuracy_tr_seed_"+str(run+training_seed)+ "_dd_seed_" + str(run_dd + data_distribution_seed) + ".txt", delimiter=',')
                                    tab_acc = tab_acc.mean(axis=1)
                                    tab_acc = tab_acc.mean(axis=0)
                                    accuracy = np.max(tab_acc)
                                    if accuracy < worst_accuracy:
                                        worst_accuracy = accuracy
                                    
                                heat_map_table[len(heat_map_table)-1-x][y] = worst_accuracy

                    heat_map_cube[z] = heat_map_table
                    
                file_name = str(
                    "best_test_"
                    + dataset_name + "_"
                    + model_name + "_"
                    + custom_dict_to_str(data_dist["name"]) + "_"
                    + pre_agg_names + ".pdf"
                )
                
                column_names = [str(dist_param) for dist_param in distribution_parameter_list]
                row_names = [str(nb_byzantine) for nb_byzantine in nb_byz]
                column_names.reverse()
                heat_map_table = np.max(heat_map_cube, axis=0)
                sns.heatmap(heat_map_table, xticklabels=row_names, yticklabels=column_names, annot=True)
                plt.savefig(path_to_plot +"/"+ file_name)
                plt.close()

find_best_hyperparameters("./results", "./best_hyperparameters")
heat_map_test_accuracy("./results", "./best_hyperparameters", "./heat_maps")
best_heat_map_test_accuracy("./results", "./best_hyperparameters", "./heat_maps")


import os
import datetime
import json

import numpy as np
import torch

class FileManager(object):
    """
    Description
    -----------
    Object whose responsability is deal with the methods and procedures to
    manage the files and store results.
    """
    def __init__(self, params=None):
        
        self.files_path = str(
            params["result_path"] + "/"
            + params["dataset_name"] + "_" 
            + params["model_name"] + "_" 
            "n_" + str(params["nb_workers"]) + "_" 
            + "f_" + str(params["nb_byz"]) + "_" 
            + "d_" + str(params["declared_nb_byz"]) + "_" 
            + params["data_distribution_name"] + "_"
            + str(params["distribution_parameter"]) + "_" 
            + params["aggregation_name"] + "_"
            + "_".join(params["pre_aggregation_names"]) + "_"
            + params["attack_name"] + "_" 
            + "lr_" + str(params["learning_rate"]) + "_" 
            + "mom_" + str(params["momentum"]) + "_" 
            + "wd_" + str(params["weight_decay"]) + "/"
        )
        
        if not os.path.exists(self.files_path):
            try:
                os.makedirs(self.files_path)
            except Exception as e:
                print(e)

        with open(self.files_path+"day.txt", "w") as file:
            file.write(str(datetime.date.today().strftime("%d_%m_%y")))
        
    def set_experiment_path(self, path):
        self.files_path = path
    
    def get_experiment_path(self):
        return self.files_path
    
    def save_config_dict(self, dict_to_save):
        with open(self.files_path+"settings.json", 'w') as json_file:
            json.dump(dict_to_save, json_file, indent=4, separators=(',', ': '))
    
    def write_array_in_file(self, array, file_name):
        np.savetxt(self.files_path+file_name, [array], fmt='%.4f', delimiter=",")
    
    def save_state_dict(self, state_dict, training_seed, data_dist_seed, step):
        if not os.path.exists(
            self.files_path+"models_tr_seed_" + str(training_seed)
            +"_dd_seed_"+str(data_dist_seed)
        ):
            os.makedirs(self.files_path+"models_tr_seed_" + str(training_seed) 
                        + "_dd_seed_"+str(data_dist_seed))
            
        torch.save(state_dict, self.files_path+"models_tr_seed_" + str(training_seed) 
                   + "_dd_seed_"+str(data_dist_seed)+"/model_step_"+ str(step) +".pth")
    
    def save_loss(self, loss_array, training_seed, data_dist_seed, client_id):
        if not os.path.exists(
            self.files_path+"loss_tr_seed_" + str(training_seed) 
            + "_dd_seed_"+str(data_dist_seed)
        ):
            os.makedirs(self.files_path + "loss_tr_seed_" + str(training_seed) 
                        + "_dd_seed_" + str(data_dist_seed))
            
        file_name = self.files_path + "loss_tr_seed_" + str(training_seed) \
                    + "_dd_seed_" + str(data_dist_seed) \
                    + "/loss_client_" + str(client_id) + ".txt"
        
        np.savetxt(file_name, loss_array, fmt='%.6f', delimiter=",")
    
    def save_accuracy(self, acc_array, training_seed, data_dist_seed, client_id):
        if not os.path.exists(
            self.files_path+"accuracy_tr_seed_" + str(training_seed) 
            + "_dd_seed_"+str(data_dist_seed)
        ):
            os.makedirs(self.files_path+"accuracy_tr_seed_" + str(training_seed) 
                        + "_dd_seed_"+str(data_dist_seed))
            
        file_name = self.files_path+"accuracy_tr_seed_" + str(training_seed) \
                    + "_dd_seed_"+str(data_dist_seed) \
                    + "/accuracy_client_" + str(client_id) + ".txt"
        
        np.savetxt(file_name, acc_array, fmt='%.4f', delimiter=",")




class ParamsManager(object):
    
    """
    Description
    -----------
    Object whose responsability is manage and store all the parameters.
    """

    def __init__(self, params):
            self.data = params
    
    def _parameter_to_use(self, default, read):
        if read is None:
            return default
        else:
            return read
    
    def _read_object(self, path):
        obj = self.data
        for idx in path:
            if idx in obj.keys():
                obj = obj[idx]
            else:
                return None
        return obj
    
    def get_flatten_info(self):
        return {
            "training_seed": self.get_training_seed(),
            "device": self.get_device(),
            "nb_workers": self.get_nb_workers(),
            "nb_honest": self.get_nb_honest(),
            "nb_byz": self.get_nb_byz(),
            "declared_nb_byz": self.get_declared_nb_byz(),
            "declared_equal_real": self.get_declared_equal_real(),
            "size_train_set": self.get_size_train_set(),
            "nb_steps": self.get_nb_steps(),
            "evaluation_delta": self.get_evaluation_delta(),
            "evaluate_on_test": self.get_evaluate_on_test(),
            "store_training_accuracy": self.get_store_training_accuracy(),
            "store_training_loss": self.get_store_training_loss(),
            "store_models": self.get_store_models(),
            "batch_size_validation": self.get_batch_size_validation(),
            "data_folder": self.get_data_folder(),
            "results_directory": self.get_results_directory(),
            "model_name": self.get_model_name(),
            "dataset_name": self.get_dataset_name(),
            "nb_labels": self.get_nb_labels(),
            "data_distribution_seed": self.get_data_distribution_seed(),
            "data_distribution_name": self.get_name_data_distribution(),
            "distribution_parameter": self.get_parameter_data_distribution(),
            "loss_name": self.get_loss(),
            "aggregator_info": self.get_aggregator_info(),
            "pre_agg_list": self.get_preaggregators(),
            "batch_norm_momentum": self.get_batch_norm_momentum(),
            "momentum": self.get_momentum(),
            "batch_size": self.get_batch_size(),
            "learning_rate": self.get_learning_rate(),
            "weight_decay": self.get_weight_decay(),
            "learning_rate_decay": self.get_learning_rate_decay(),
            "milestones": self.get_milestones(),
            "attack_name": self.get_attack_name(),
            "attack_parameters": self.get_attack_parameters(),
            "attack_optimizer_name": self.get_attack_optimizer_name(),
            "attack_optimizer_parameters": self.get_attack_optimizer_parameters()
        }
    
    def get_data(self):
        return {   
            "general": {
                "training_seed": self.get_training_seed(),
                "device": self.get_device(),
                "nb_workers": self.get_nb_workers(),
                "nb_honest": self.get_nb_honest(),
                "nb_byz": self.get_nb_byz(),
                "declared_nb_byz": self.get_declared_nb_byz(),
                "declared_equal_real": self.get_declared_equal_real(),
                "size_train_set": self.get_size_train_set(),
                "nb_steps": self.get_nb_steps(),
                "evaluation_delta": self.get_evaluation_delta(),
                "evaluate_on_test": self.get_evaluate_on_test(),
                "store_training_accuracy": self.get_store_training_accuracy(),
                "store_training_loss": self.get_store_training_loss(),
                "store_models": self.get_store_models(),
                "batch_size_validation": self.get_batch_size_validation(),
                "data_folder": self.get_data_folder(),
                "results_directory": self.get_results_directory(),
            },

            "model": {
                "name": self.get_model_name(),
                "dataset_name": self.get_dataset_name(),
                "nb_labels": self.get_nb_labels(),
                "data_distribution_seed": self.get_data_distribution_seed(),
                "data_distribution": {
                        "name": self.get_name_data_distribution(),
                        "distribution_parameter": self.get_parameter_data_distribution()
                },
                "loss": self.get_loss(),
            },

            "aggregator": {
                    "name": self.get_aggregator_name(),
                    "parameters": self.get_aggregator_parameters()
            },
        
            "pre_aggregators" : self.get_preaggregators(),

            "server": {
                "batch_norm_momentum": self.get_batch_norm_momentum(),
                "learning_rate_decay": self.get_learning_rate_decay(),
                "milestones": self.get_milestones()
            },

            "honest_nodes": {
                "momentum": self.get_momentum(),
                "batch_size": self.get_batch_size(),
                "learning_rate": self.get_learning_rate(),
                "weight_decay": self.get_weight_decay()
            },

            "attack": {
                "name": self.get_attack_name(),
                "parameters": self.get_attack_parameters(),
                "attack_optimizer": {
                    "name": self.get_attack_optimizer_name(),
                    "parameters": self.get_attack_optimizer_parameters()
                }
            }
        }

    def get_training_seed(self):
        default = 0
        path = ["general", "training_seed"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_device(self):
        default = "cpu"
        path = ["general", "device"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_nb_workers(self):
        default = 15
        path = ["general", "nb_workers"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_nb_honest(self):
        default = None
        path = ["general", "nb_honest"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_nb_byz(self):
        default = 0
        path = ["general", "nb_byz"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_declared_nb_byz(self):
        default = 0
        path = ["general", "declared_nb_byz"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_declared_equal_real(self):
        default = False
        path = ["general", "declared_equal_real"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_size_train_set(self):
        default = 0.8
        path = ["general", "size_train_set"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_nb_steps(self):
        default = 1000
        path = ["general", "nb_steps"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_evaluation_delta(self):
        default = 50
        path = ["general", "evaluation_delta"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_evaluate_on_test(self):
        default = True
        path = ["general", "evaluate_on_test"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)

    def get_store_training_accuracy(self):
        default = True
        path = ["general", "store_training_accuracy"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_store_training_loss(self):
        default = True
        path = ["general", "store_training_loss"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_store_models(self):
        default = True
        path = ["general", "store_models"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)

    def get_batch_size_validation(self):
        default = 100
        path = ["general", "batch_size_validation"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_data_folder(self):
        default = "./data"
        path = ["general", "data_folder"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_results_directory(self):
        default = "./results"
        path = ["general", "results_directory"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_model_name(self):
        default = "cnn_mnist"
        path = ["model", "name"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)

    def get_dataset_name(self):
        default = "mnist"
        path = ["model", "dataset_name"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_nb_labels(self):
        path = ["model", "nb_labels"]
        return self._read_object(path)

    def get_data_distribution_seed(self):
        default = 0
        path = ["model", "data_distribution_seed"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_name_data_distribution(self):
        default = "iid"
        path = ["model", "data_distribution", "name"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_parameter_data_distribution(self):
        default = None
        path = ["model", "data_distribution", "distribution_parameter"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_loss(self):
        default = "NLLLoss"
        path = ["model", "loss"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_aggregator_info(self):
        default = {"name": "Average", "parameters": {}}
        path = ["aggregator"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_aggregator_name(self):
        default = "average"
        path = ["aggregator", "name"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_aggregator_parameters(self):
        path = ["aggregator", "parameters"]
        return self._read_object(path)
    
    def get_preaggregators(self):
        path = ["pre_aggregators"]
        return self._read_object(path)

    def get_batch_norm_momentum(self):
        default = 0.1
        path = ["server", "batch_norm_momentum"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)

    def get_learning_rate_decay(self):
        default = 5000
        path = ["server", "learning_rate_decay"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_momentum(self):
        default = 0.99
        path = ["honest_nodes", "momentum"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_batch_size(self):
        default = 25
        path = ["honest_nodes", "batch_size"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)

    def get_learning_rate(self):
        default = 0.1
        path = ["honest_nodes", "learning_rate"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_weight_decay(self):
        default = 0
        path = ["honest_nodes", "weight_decay"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)

    def get_milestones(self):
        default = 200
        path = ["server", "milestones"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_attack_name(self):
        default = "no_attack"
        path = ["attack", "name"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_attack_parameters(self):
        default = "NoAttack"
        path = ["attack", "parameters"]
        read = self._read_object(path)
        return self._parameter_to_use(default, read)
    
    def get_attack_optimizer_name(self):
        default = None
        if "attack_optimizer" in self.data["attack"]:
            if "name" in self.data["attack"]["attack_optimizer"]:
                path = ["attack", "attack_optimizer", "name"]
                read = self._read_object(path)
                return self._parameter_to_use(default, read)
        return default
    
    def get_attack_optimizer_parameters(self):
        default = {}
        if "attack_optimizer" in self.data["attack"]:
            if "parameters" in self.data["attack"]["attack_optimizer"]:
                path = ["attack", "attack_optimizer", "parameters"]
                read = self._read_object(path)
                return self._parameter_to_use(default, read)
        return default



import argparse
import json
from multiprocessing import Pool, Value
import os
import copy

from byzfl.pipeline.train import Train

default_settings = {
    "general": {
        "training_seed": 0,
        "nb_training_seeds": 5,
        "device": "cuda",
        "nb_workers": None,
        "nb_honest": 10,
        "nb_byz": [1, 3, 5, 7, 9],
        "declared_nb_byz": [1, 3, 5, 7, 9],
        "declared_equal_real": True,
        "size_train_set": 0.8,
        "nb_steps": 800,
        "evaluation_delta": 50,
        "evaluate_on_test": True,
        "store_training_accuracy": True,
        "store_training_loss": True,
        "store_models": False,
        "data_folder": None,
        "results_directory": "results"
    },
    "model": {
        "name": "cnn_mnist",
        "dataset_name": "mnist",
        "nb_labels": 10,
        "data_distribution_seed": 0,
        "nb_data_distribution_seeds": 1,
        "data_distribution": [
            {
                "name": "gamma_similarity_niid",
                "distribution_parameter": [1.0, 0.75, 0.5, 0.25, 0.0]
            }
        ],
        "loss": "NLLLoss"
    },
    "aggregator": [
        {
            "name": "Median",
            "parameters": {}
        },
        {
            "name": "TrMean",
            "parameters": {}
        },
        {
            "name": "GeometricMedian",
            "parameters": {
                "nu": 0.1,
                "T": 3
            }
        },
        {
            "name": "MultiKrum",
            "parameters": {}
        }
    ],
    "pre_aggregators": [
        {
            "name": "Clipping", 
            "parameters": {}
        },
        {
            "name": "NNM", 
            "parameters": {}
        }
    ],
    "server": {
        "batch_norm_momentum": None,
        "batch_size_validation": 100,
        "learning_rate_decay": 1.0,
        "milestones": []
    },
    "honest_nodes": {
        "momentum": [0.0, 0.25, 0.5, 0.75, 0.9, 0.99],
        "batch_size": 25,
        "learning_rate": [0.1, 0.25, 0.35],
        "weight_decay": 1e-4
    },
    "attack": [
        {
            "name": "SignFlipping",
            "parameters": {},
            "attack_optimizer": {
                "name": None
            }
        },
        {
            "name": "LabelFlipping",
            "parameters": {},
            "attack_optimizer": {
                "name": None
            }
        },
        {
            "name": "FallOfEmpires",
            "parameters": {},
            "attack_optimizer": {
                "name": "LineMaximize"
            }
        },
        {
            "name": "LittleIsEnough",
            "parameters": {},
            "attack_optimizer": {
                "name": "LineMaximize"
            }
        },
        {
            "name": "Mimic",
            "parameters": {},
            "attack_optimizer": {
                "name": "WorkerWithMaxVariance",
                "parameters": {
                    "steps_to_learn": 200
                }
            }
        }
    ]
}

def generate_all_combinations_aux(list_dict, orig_dict, aux_dict, rest_list):
    if len(aux_dict) < len(orig_dict):
        key = list(orig_dict)[len(aux_dict)]
        if isinstance(orig_dict[key], list):
            if not orig_dict[key] or (key in rest_list and 
                not isinstance(orig_dict[key][0], list)):
                aux_dict[key] = orig_dict[key]
                generate_all_combinations_aux(list_dict, 
                                              orig_dict, 
                                              aux_dict, 
                                              rest_list)
            else:
                for item in orig_dict[key]:
                    if isinstance(item, dict):
                        new_list_dict = []
                        new_aux_dict = {}
                        generate_all_combinations_aux(new_list_dict, 
                                                    item, 
                                                    new_aux_dict, 
                                                    rest_list)
                    else:
                        new_list_dict = [item]
                    for new_dict in new_list_dict:
                        new_aux_dict = aux_dict.copy()
                        new_aux_dict[key] = new_dict
                        
                        generate_all_combinations_aux(list_dict,
                                                    orig_dict, 
                                                    new_aux_dict, 
                                                    rest_list)
        elif isinstance(orig_dict[key], dict):
            new_list_dict = []
            new_aux_dict = {}
            generate_all_combinations_aux(new_list_dict, 
                                          orig_dict[key], 
                                          new_aux_dict, 
                                          rest_list)
            for dictionary in new_list_dict:
                new_aux_dict = aux_dict.copy()
                new_aux_dict[key] = dictionary
                generate_all_combinations_aux(list_dict, 
                                              orig_dict, 
                                              new_aux_dict, 
                                              rest_list)
        else:
            aux_dict[key] = orig_dict[key]
            generate_all_combinations_aux(list_dict, 
                                          orig_dict, 
                                          aux_dict, 
                                          rest_list)
    else:
        list_dict.append(aux_dict)

def generate_all_combinations(original_dict, restriction_list):
    list_dict = []
    aux_dict = {}
    generate_all_combinations_aux(list_dict, original_dict, aux_dict, restriction_list)
    return list_dict

def init_pool_processes(shared_value):
    global counter
    counter = shared_value

def run_training(params):
    train = Train(params)
    train.run_SGD()
    with counter.get_lock():
        print("Training " + str(counter.value) + " done")
        counter.value += 1

def process_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('--nb_jobs', type=int, help='Number of Jobs (multiprocessing)')

    args = parser.parse_args()
    
    nb_jobs = 1
    if args.nb_jobs is not None:
        nb_jobs = args.nb_jobs
    print("Running " + str(nb_jobs) + " experiments in parallel")
    return nb_jobs

def eliminate_experiments_done(dict_list):
    if len(dict_list) != 0:
        directory = dict_list[0]["general"]["results_directory"]
        folders = [name for name in os.listdir(directory) if os.path.isdir(os.path.join(directory, name))]
        if len(folders) != 0:
            real_dict_list = []
            for setting in dict_list:
                if setting["general"]["nb_workers"] == None:
                    setting["general"]["nb_workers"] = setting["general"]["nb_honest"] + setting["general"]["nb_byz"]
                # First check folder
                pre_aggregation_names =  [
                    dict['name']
                    for dict in setting["pre_aggregators"]
                ]
                folder_name = str(
                    setting["model"]["dataset_name"] + "_" 
                    + setting["model"]["name"] + "_" 
                    +"n_" + str(setting["general"]["nb_workers"]) + "_" 
                    + "f_" + str(setting["general"]["nb_byz"]) + "_" 
                    + "d_" + str(setting["general"]["declared_nb_byz"]) + "_"
                    + setting["model"]["data_distribution"]["name"] + "_"
                    + str(setting["model"]["data_distribution"]["distribution_parameter"]) + "_" 
                    + setting["aggregator"]["name"] + "_"
                    + "_".join(pre_aggregation_names) + "_"
                    + setting["attack"]["name"] + "_" 
                    + "lr_" + str(setting["honest_nodes"]["learning_rate"]) + "_" 
                    + "mom_" + str(setting["honest_nodes"]["momentum"]) + "_" 
                    + "wd_" + str(setting["honest_nodes"]["weight_decay"])
                )

                if folder_name in folders:
                    #Now we check the seeds
                    training_seed = setting["general"]["training_seed"]
                    data_distribution_seed = setting["model"]["data_distribution_seed"]
                    files = os.listdir(directory+"/"+folder_name)
                    name = "train_time_tr_seed_" + str(training_seed) + "_dd_seed_" + str(data_distribution_seed) + ".txt"
                    if not name in files:
                        real_dict_list.append(setting)
                else:
                    real_dict_list.append(setting)
            return real_dict_list
        else:
            return dict_list
    else:
        return dict_list

def delegate_training_seeds(dict_list):
    real_dict_list = []
    for setting in dict_list:
        original_seed = setting["general"]["training_seed"]
        nb_seeds = setting["general"]["nb_training_seeds"]
        for i in range(nb_seeds):
            new_setting = copy.deepcopy(setting)
            new_setting["general"]["training_seed"] = original_seed + i
            real_dict_list.append(new_setting)
    return real_dict_list

def delegate_data_distribution_seeds(dict_list):
    real_dict_list = []
    for setting in dict_list:
        original_seed = setting["model"]["data_distribution_seed"]
        nb_seeds = setting["model"]["nb_data_distribution_seeds"]
        for i in range(nb_seeds):
            new_setting = copy.deepcopy(setting)
            new_setting["model"]["data_distribution_seed"] = original_seed + i
            real_dict_list.append(new_setting)
    return real_dict_list

def remove_real_greater_declared(dict_list):
    real_dict_list = []
    for setting in dict_list:
        if setting["general"]["declared_nb_byz"] >= setting["general"]["nb_byz"]:
            real_dict_list.append(setting)
    return real_dict_list

def remove_real_not_equal_declared(dict_list):
    real_dict_list = []
    for setting in dict_list:
        if setting["general"]["declared_nb_byz"] == setting["general"]["nb_byz"]:
            real_dict_list.append(setting)
    return real_dict_list

"""
if __name__ == '__main__':
    nb_jobs = process_args()
    data = {}
    try:
        with open('settings.json', 'r') as file:
            data = json.load(file)
    except:
        print(f"{'settings.json'} not found.")

        with open('settings.json', 'w') as f:
            json.dump(default_settings, f, indent=4)

        print(f"{'settings.json'} created successfully.")
        print("Please configure the experiment you want to run.")
        exit()
    
    results_directory = None
    if data["general"]["results_directory"] is None:
        results_directory = "./results"
    else:
        results_directory = data["general"]["results_directory"]

    if not os.path.exists(results_directory):
        os.makedirs(results_directory)
    
    with open(results_directory+"/settings.json", 'w') as json_file:
            json.dump(data, json_file, indent=4, separators=(',', ': '))

    restriction_list = ["pre_aggregators", "milestones"]
    dict_list = generate_all_combinations(data, restriction_list)

    if data["general"]["declared_equal_real"]:
        dict_list = remove_real_not_equal_declared(dict_list)
    else:
        dict_list = remove_real_greater_declared(dict_list)

    #Do a setting for every seed
    dict_list = delegate_training_seeds(dict_list)
    dict_list = delegate_data_distribution_seeds(dict_list)

    #Do only experiments that haven't been done
    dict_list = eliminate_experiments_done(dict_list)

    print("Total trainings to do: " + str(len(dict_list)))

    counter = Value('i', 0)
    with Pool(initializer=init_pool_processes, initargs=(counter,), processes=nb_jobs) as pool:
        pool.map(run_training, dict_list)
"""

def run_experiment(nb_jobs=1):
    data = {}
    try:
        with open('settings.json', 'r') as file:
            data = json.load(file)
    except:
        print(f"{'settings.json'} not found.")

        with open('settings.json', 'w') as f:
            json.dump(default_settings, f, indent=4)

        print(f"{'settings.json'} created successfully.")
        print("Please configure the experiment you want to run.")
        exit()
    
    results_directory = None
    if data["general"]["results_directory"] is None:
        results_directory = "./results"
    else:
        results_directory = data["general"]["results_directory"]

    if not os.path.exists(results_directory):
        os.makedirs(results_directory)
    
    with open(results_directory+"/settings.json", 'w') as json_file:
            json.dump(data, json_file, indent=4, separators=(',', ': '))

    restriction_list = ["pre_aggregators", "milestones"]
    dict_list = generate_all_combinations(data, restriction_list)

    if data["general"]["declared_equal_real"]:
        dict_list = remove_real_not_equal_declared(dict_list)
    else:
        dict_list = remove_real_greater_declared(dict_list)

    #Do a setting for every seed
    dict_list = delegate_training_seeds(dict_list)
    dict_list = delegate_data_distribution_seeds(dict_list)

    #Do only experiments that haven't been done
    dict_list = eliminate_experiments_done(dict_list)

    print("Total trainings to do: " + str(len(dict_list)))

    counter = Value('i', 0)
    with Pool(initializer=init_pool_processes, initargs=(counter,), processes=nb_jobs) as pool:
        pool.map(run_training, dict_list)
